/*++ nqs_quereq.c - Network Queueing System
 *
 * $Source$
 *
 * DESCRIPTION:
 *
 *	Queue an NQS request for execution.
 *
 *
 *	Author:
 *	-------
 *	Brent A. Kingsbury, Sterling Software Incorporated.
 *	August 12, 1985.
 *
 *
 * STANDARDS VIOLATIONS:
 *   None.
 *
 * REVISION HISTORY: ($Revision$ $Date$ $State$)
 * $Log$
 *
 *	October 9, 1991	Move requests between queues
 *	October 15, 1991 Hold and release requests
 *	October 29, 1992 Rerun requests
 *
 */

#include <stdio.h>
#include <signal.h>
#if	SGI | SYS52 | SYS53 | UNICOS | UTS
#include <fcntl.h>			/* File control */
#else
#if	BSD42 | BSD43 | ULTRIX
#include <sys/file.h>
#else
BAD SYSTEM TYPE
#endif
#endif
#include "nqs.h"			/* NQS constants and data types */
#include "informcc.h"			/* NQS information completion */
					/* codes and masks */
#include "transactcc.h"			/* NQS transaction completion codes */
#include "nqsxvars.h"			/* NQS global vars and directories */

extern char *asciierrno();		/* Return ASCII errno */
extern int getreq();			/* Open/read request header */
extern void mkdefault();		/* Make default output file name */
extern void nqs_delrfs();		/* Delete the specified control and */
					/* data request files */
extern int nqs_enfile();		/* Diagnose open error returning */
					/* non-zero if too many file */
					/* descriptors were open causing */
					/* the last open to fail; otherwise */
					/* 0 is returned. */
extern long nsq_enque();		/* Insert a request into a queue */
extern struct qhead *nqs_fndnnq();	/* Find non-network queue */
extern struct request *nqs_fndreq();	/* Find request by sequence# & mid */
extern void nsq_unque();		/* Unqueue the most recently queued */
					/* request */
extern void pack6name();		/* Compute name on 6-bit alphabet */
extern int readreq();			/* Read the request header portion */
					/* of a control file for an NQS req */
extern int strcmp();			/* String comparison */
extern char *strcpy();			/* String copy */
extern int strlen();			/* String length */
extern time_t time();			/* Time */
extern void udb_useqno();		/* Update next available user req */
					/* sequence number stored in the */
					/* NQS database. */
extern int writereq();			/* Write/rewrite request header */

#define MDELTA 15			/* Mimimum time interval */

/*** nqs_quereq
 *
 *
 *	long nqs_quereq():
 *	Queue an NQS request for execution.
 *
 *	Returns:
 *		TCML_SUBMITTED: if successful in queueing the request;
 *		otherwise, a failure transaction code is returned (TCMx_).
 *		The information completion code portion of the return
 *		value is 0.
 */
long nqs_quereq (mode, ctrlname, orig_seqno, orig_mid, destqueue, frommid)
int mode;				/* Queueing mode:  0=new; 1=local */
					/* pipe queue; 2=remote pipe queue */
char *ctrlname;				/* The name of the req control file */
					/* (Only meaningful if mode = 0) */
long orig_seqno;			/* Original request sequence# */
					/* (Only meaningful if mode > 0) */
mid_t orig_mid;				/* Original request machine-id */
					/* (Only meaningful if mode > 0) */
char *destqueue;			/* Destination queue */
					/* (Only meaningful if mode > 0) */
mid_t frommid;				/* Machine delivering request */
{					/* (Only meaningful if mode = 2) */

	char newpath [MAX_PATHNAME+1];	/* New pathname */
	char oldpath [MAX_PATHNAME+1];	/* Old pathname */
	struct rawreq rawreq;		/* Raw request structure */
	int cfd;			/* Control file file-descriptor */
	long insqueres;			/* nsq_enque() result code */
	register int i;			/* Counter */

	if (Debug) {
		printf ("D$entering nqs_quereq().\n");
		fflush (stdout);
	}
	if (mode == 0) {
		/*
		 *  This request is a brand new request, with its
		 *  associated files in the new local request
		 *  directory.
		 */
		sprintf (newpath, "%s/%s", Nqs_requests, ctrlname);
	}
	else {
		/*
		 *  The request is being routed by a pipe
		 *  queue.
		 */
		pack6name (newpath, Nqs_control,
			  (int) (orig_seqno % MAX_CTRLSUBDIRS), (char *) 0,
			  (long) orig_seqno, 5, (long) orig_mid, 6, 0, 0);
	}
	if ((cfd = open (newpath, O_RDWR)) == -1) {
		/*
		 *  We are unable to open the control file!
		 */
		if (nqs_enfile ()) {
			printf ("I$Unable to accept new request because of ");
			printf ("file descriptor shortage.\n");
			fflush (stdout);
			return (TCML_ENFILE);
		}
		return (TCML_UNAFAILURE);
	}
	if (Debug) {
		printf ("D$Attempting to read request.\n");
		fflush (stdout);
	}
	if (readreq (cfd, &rawreq) == -1) {
		/*
		 *  Request is corrupt.
		 */
		if (mode == 0) {
			rawreq.tcm = TCML_UNAFAILURE;
			writereq (cfd, &rawreq);
		}
		close (cfd);			/* Close request file */
		return (TCML_UNAFAILURE);	/* Failure */
	}
	if (Debug) {
		printf ("D$Request was read.\n");
		fflush (stdout);
	}
	/*
	 *  Assign sequence number to request.
	 */
	if (mode == 0) {
		/*
		 *  The request is being submitted for the first time and
		 *  does not have a sequence number or transaction-id
		 *  assigned to it.
		 */
		rawreq.trans_id = 0;		/* No arguments about this */
		time (&rawreq.create_time);
		rawreq.tcm = TCML_SUBMITTED;	/* Might overwrite this below */
		rawreq.orig_seqno = Seqno_user;	/* Seq# the request will have */
	}					/* if it queues successfully */
	else {
		/*
		 *  The request is being routed by a pipe queue.
		 *  Copy the destination queue name into the queue
		 *  field of the rawreq structure for the request.
		 */
		strcpy (rawreq.quename, destqueue);
		if (mode == 2) {
			/*
			 *  Set values out of paranoia (that should
			 *  already be set) for requests that are
			 *  being queued from a remote machine.
			 */
			rawreq.trans_id = 0;	/* No transaction-id */
			rawreq.flags |= RQF_EXTERNAL;
		}
	}
	/*
	 *  Assign queue entry time to request.
	 */
	time (&rawreq.enter_time);
	/*
	 *  Manufacture default output file pathnames for batch request
	 *  stderr, stdlog, and stdout files as necessary.
	 */
	if (rawreq.type == RTYPE_BATCH) {
		i = strlen (rawreq.v.bat.stderr_name);
		if ((rawreq.v.bat.stderr_acc & OMD_EO) == 0 &&
		    (i == 0 || rawreq.v.bat.stderr_name [i-1] == '/')) {
			/*
			 *  The stderr file name needs to be resolved.
			 *  The user did NOT specify the pathname for
			 *  the stderr file, and the stderr file is to
			 *  be placed on the execution machine, or
			 *  returned to the submitting machine as
			 *  determined by the OMD_M_KEEP modifier bit.
			 */
			if (Debug > 1) {
				printf ("D$Stderr name needs resolution.\n");
				fflush (stdout);
			}
			if (resolvepath (&rawreq, &rawreq.v.bat.stderr_mid,
					 rawreq.v.bat.stderr_name,
					 rawreq.v.bat.stderr_acc, i,
					 'e') == -1) {
				/*
				 *  The resolved pathname will be too long.
				 */
				if (mode == 0) {
					rawreq.tcm = TCML_PATHLEN;
					writereq (cfd, &rawreq);
				}
				close (cfd);		/* Close request file */
				return (TCML_PATHLEN);
			}
			if (Debug > 1) {
				printf ("D$Resolved stderr name = %s.\n",
					 rawreq.v.bat.stderr_name);
				fflush (stdout);
			}
		}
		i = strlen (rawreq.v.bat.stdlog_name);
		if (i == 0 || rawreq.v.bat.stdlog_name [i-1] == '/') {
			/*
			 *  The stdlog file name needs to be resolved.
			 *  The user did NOT specify the pathname for
			 *  the stdlog file, and the stdlog file is to
			 *  be placed on the execution machine, or
			 *  returned to the submitting machine as
			 *  determined by the OMD_M_KEEP modifier bit.
			 */
			if (Debug > 1) {
				printf ("D$Stdlog name needs resolution.\n");
				fflush (stdout);
			}
			if (resolvepath (&rawreq, &rawreq.v.bat.stdlog_mid,
					 rawreq.v.bat.stdlog_name,
					 rawreq.v.bat.stdlog_acc, i,
					 'l') == -1) {
				/*
				 *  The resolved pathname will be too long.
				 */
				if (mode == 0) {
					rawreq.tcm = TCML_PATHLEN;
					writereq (cfd, &rawreq);
				}
				close (cfd);		/* Close request file */
				return (TCML_PATHLEN);
			}
			if (Debug > 1) {
				printf ("D$Resolved stdlog name = %s.\n",
					 rawreq.v.bat.stdlog_name);
				fflush (stdout);
			}
		}
		i = strlen (rawreq.v.bat.stdout_name);
		if (i == 0 || rawreq.v.bat.stdout_name [i-1] == '/') {
			/*
			 *  The stdout file name needs to be resolved.
			 *  The user did NOT specify the pathname for
			 *  the stdout file, and the stdout file is to
			 *  be placed on the execution machine, or
			 *  returned to the submitting machine as
			 *  determined by the OMD_M_KEEP modifier bit.
			 */
			if (Debug > 1) {
				printf ("D$Stdout name needs resolution.\n");
				fflush (stdout);
			}
			if (resolvepath (&rawreq, &rawreq.v.bat.stdout_mid,
					 rawreq.v.bat.stdout_name,
					 rawreq.v.bat.stdout_acc, i,
					 'o') == -1) {
				/*
				 *  The resolved pathname will be too long.
				 */
				if (mode == 0) {
					rawreq.tcm = TCML_PATHLEN;
					writereq (cfd, &rawreq);
				}
				close (cfd);		/* Close request file */
				return (TCML_PATHLEN);
			}
			if (Debug > 1) {
				printf ("D$Resolved stdout name = %s.\n",
					 rawreq.v.bat.stdout_name);
				fflush (stdout);
			}
		}
	}
	/*
	 *  The request has been assigned a sequence number, and any
	 *  unresolved output return pathnames have also been resolved.
	 *
	 *  Try to insert the request into the internal queues of NQS.
	 */
	insqueres = nsq_enque (mode, cfd, &rawreq, 0, frommid,
			       rawreq.enter_time);
	if ((insqueres & XCI_FULREA_MASK) != TCML_SUBMITTED) {
		/*
		 *  The request was NOT submitted.
		 */
		if (mode == 0) {
			rawreq.tcm = insqueres;	/* Store transaction and */
						/* information completion */
						/* codes */
			writereq (cfd, &rawreq);/* Update request header */
		}
		close (cfd);			/* Close request file */
		return (insqueres);		/* Return transaction code */
	}
	/*
	 *  The request was successfully enqueued.  Make sure to increment
	 *  the next available request sequence number, if necessary.
	 */
	if (mode == 0) {
		rawreq.tcm = insqueres;	/* Store transaction and information */
					/* completion codes */
		/*
		 *  Update the next available sequence number.
		 */
		if (++Seqno_user > MAX_SEQNO_USER) Seqno_user = 0;
		udb_useqno (Seqno_user);	/* Update */
		/*
		 *  A Sequence number and transaction-id has been assigned
		 *  to the request, and the request has been successfully
		 *  enqueued.  Form links in the NQS data directory to all
		 *  data files that are a part of the request.
		 *
		 *  Note however that the control file has NOT been updated
		 *  with the assigned sequence number or transaction-id.  If
		 *  the system crashes after the links have been made, but
		 *  with the control file un-updated, then nqs_rbuild.c will
		 *  have to catch it later.
		 */
		if (Debug) {
			printf ("D$Trying to link in files.\n");
			fflush (stdout);
		}
		for (i=0; i < rawreq.ndatafiles; i++) {
			pack6name (newpath, Nqs_data,
				  (int) (rawreq.orig_seqno % MAX_DATASUBDIRS),
				  (char *) 0, (long) rawreq.orig_seqno, 5,
				  (long) rawreq.orig_mid, 6, i, 3);
			pack6name (oldpath, Nqs_requests, -1, ctrlname, 0L, 0,
				   0L, 0, i, 3);
			if (link (oldpath, newpath) == -1) {
				/*
				 *  The link failed!
				 */
				printf ("E$Unable to form link to ");
				printf ("new request data file.\n");
				printf ("I$link (%s, %s) failed;\n",
					 oldpath, newpath);
				printf ("I$%s.\n", asciierrno());
				fflush (stdout);
				/*
				 *  Unlink any previous links to data files
				 *  for the new request.
				 */
				nqs_delrfs (rawreq.orig_seqno,
					    rawreq.orig_mid, i);
				rawreq.tcm = TCML_UNAFAILURE;
				writereq (cfd, &rawreq);
				close (cfd);	/* Close request file */
				nsq_unque();	/* Unqueue the request */
				return (TCML_UNAFAILURE);	/* Failure */
			}
			unlink (oldpath);	/* Unlink original */
		}
		/*
		 *  Form the link to the control file for the new request.
		 */
		pack6name (newpath, Nqs_control,
			  (int) (rawreq.orig_seqno % MAX_CTRLSUBDIRS),
			  (char *) 0, (long) rawreq.orig_seqno, 5,
			  (long) rawreq.orig_mid, 6, 0, 0);
		sprintf (oldpath, "%s/%s", Nqs_requests, ctrlname);
		if (link (oldpath, newpath) == -1) {
			/*
			 *  We failed.
			 */
			printf ("E$Unable to form link to new request ");
			printf ("control file.\n");
			printf ("I$link (%s, %s) failed;\n", oldpath, newpath);
			printf ("I$%s.\n", asciierrno());
			fflush (stdout);
			/*
			 *  Unlink all of the data files and control files
			 *  for the new request.
			 */
			nqs_delrfs (rawreq.orig_seqno, rawreq.orig_mid,
				    rawreq.ndatafiles);
			rawreq.tcm = TCML_UNAFAILURE;
			writereq (cfd, &rawreq);
			close (cfd);		/* Close request file */
			nsq_unque();		/* Unqueue the request */
			return (TCML_UNAFAILURE);
		}				/* Failure */
		unlink (oldpath);		/* Unlink original */
	}
	/*
	 *  The request has been assigned a sequence number and a transaction
	 *  descriptor.  The request has also been queued, and all of the
	 *  control and data files have been linked in.  Rawreq.tcm holds
	 *  the transaction completion code.
	 */
	writereq (cfd, &rawreq);		/* Update request header */
	close (cfd);				/* Close request file */
	return (insqueres);			/* Return transaction code */
}


/*** resolvepath
 *
 *
 *	int resolvepath():
 *	Resolve batch request output file path (when necessary).
 *
 *	Returns:
 *		 0: Path resolved successfully.
 *		-1: Resolved path is too long.
 */
static int resolvepath (rawreq, mid_resolve, path_resolve, access_mode, length,
			suffixchar)
register struct rawreq *rawreq;		/* Raw request structure for request */
mid_t *mid_resolve;			/* Machine-id requiring resolution */
register char *path_resolve;		/* Path requiring resolution */
short access_mode;			/* Access mode */
register int length;			/* Len of path requiring resolution */
char suffixchar;			/* Suffix char for default */
{
	char dirpath [MAX_REQPATH+1];	/* Output file pathname */

	if ((length + MAX_REQNAME + 7) > MAX_REQPATH) {
		/*
		 *  The resolved pathname will be
		 *  too long.
		 */
		rawreq->tcm = TCML_PATHLEN;
		return (-1);		/* Too long */
	}
	strcpy (dirpath, path_resolve);
	if (length) dirpath [length-1] = '\0';
	/*
	 *  Manufacture proper default pathname.
	 */
	mkdefault (path_resolve, dirpath, rawreq->reqname,
		  (long) rawreq->orig_seqno, suffixchar);
	if (access_mode & OMD_M_KEEP) {
		/*
		 *  The file must be left on the execution machine.
		 */
		*mid_resolve = Locmid;
	}
	return (0);			/* Success */
}


/*** nqs_movreq
 *
 *
 *	long nqs_movreq():
 *	Move a queued NQS request.
 *
 *	Returns:
 *		TCML_COMPLETE:	if the request was successfully modified.
 *		TCML_NOSUCHQUE:	if the specified queue does not existed
 *				on this machine.
 *		TCML_QUEDISABL:	if the specified queue is disabled.
 *		TCML_NOSUCHREQ:	if the specified request does not exist
 *				on this machine.
 *		TCML_WROQUETYP:	if the destination queue type is not
 *				compatible with the source queue type.
 */
long nqs_movreq (orig_seqno, orig_mid, destque)
long orig_seqno;		/* Originating sequence number of request */
mid_t orig_mid;			/* Originating machine id */
char *destque;			/* Destination queue name */
{
	register struct qhead *queue;	/* Queue in which req is placed */
	long rc;
	long nqs_move();

	if (Debug > 2) {
		printf ("D$nqs_movreq(): seqno=%d; mid=%d; queue=%s\n",
			orig_seqno, orig_mid, destque);
		fflush (stdout);
	}
	/*
	 *  Locate the queue in which to place the request.
	 */
	queue = nqs_fndnnq (destque);
	if (queue == (struct qhead *) 0) {
		return (TCML_NOSUCHQUE);	/* No such queue */
	}
	if (!(queue->q.status & QUE_ENABLED)) {
		return (TCML_QUEDISABL);	/* Queue is disabled */
	}
	rc = nqs_move (orig_seqno, orig_mid, queue);
	if (rc != TCML_COMPLETE) return (rc);
	switch (queue->q.type) {
	case QUE_BATCH:
		bsc_spawn();		/* Maybe spawn some batch reqs */
		break;
	case QUE_DEVICE:
		dsc_spawn();		/* Maybe spawn some device reqs */
		break;
	case QUE_PIPE:
		psc_spawn();		/* Maybe spawn some pipe reqs */
		break;
	}
	if (queue->q.status & QUE_UPDATE) {
		/*
		 *  No requests were spawned from the queue in which
		 *  the most recent request was placed.
		 */
		udb_qorder (queue);	/* Update and clear QUE_UPDATE bit */
	}
	return (TCML_COMPLETE);		/* Return transaction code */
}


/*** nqs_movque
 *
 *
 *	long nqs_movque():
 *	Move all of the NQS requests from one queue to another.
 *
 *	Returns:
 *		TCML_COMPLETE:	if the request was successfully modified.
 *		TCML_NOSUCHQUE:	if the specified queue does not exist
 *				on this machine.
 *		TCML_QUEDISABL:	if the specified queue is disabled.
 *		TCML_NOSUCHREQ:	if the specified request does not exist
 *				on this machine.
 *		TCML_WROQUETYP:	if the destination queue type is not
 *				compatible with the source queue type.
 */
long nqs_movque (from_que, to_que)
char *from_que;				/* Queue to move requests from */
char *to_que;				/* Queue to move requests to */
{
	register struct qhead *from_queue;  /* Queue to move requests from */
	register struct qhead *to_queue;    /* Queue to move requests from */
	register struct request *req;	/* Request struct allocated for req */
	register struct request *next;	/* Next request */
	register long rc;		/* return code from nqs_movreq() */
	long nqs_movqtyp();

	/*
	 *  Locate the queue from which requests will be moved.
	 */
	from_queue = nqs_fndnnq (from_que);
	if (from_queue == (struct qhead *) 0) {
		return (TCML_NOSUCHQUE);	/* No such queue */
	}
	if (Debug > 2) {
		printf ("D$nqs_movque(): Attempting to move requests ");
		printf ("from queue %s to queue %s\n",
			from_que, to_que);
		fflush (stdout);
	}
	/*
	 *  Locate the queue to which requests will be moved.
	 */
	to_queue = nqs_fndnnq (to_que);
	if (to_queue == (struct qhead *) 0)
		return (TCML_NOSUCHQUE);	/* No such queue */
	if (!(to_queue->q.status & QUE_ENABLED))
		return (TCML_QUEDISABL);	/* Queue is disabled */
	if (strcmp(from_que, to_que) == 0)	/* if same queue */
		return (TCML_COMPLETE);
	if (Debug > 2) {
		printf ("D$nqs_movque(): Attempting to move requests ");
		printf ("to queue %s\n", to_que);
		fflush (stdout);
	}
	/*
	 *  Move all requests to the destination queue except for those
	 *  that are departing, arriving, or running.
	 */
	/* if ((rc = nqs_movqtyp (from_queue->runset, to_queue)) !=
		TCML_COMPLETE) return (rc); */
	if ((rc = nqs_movqtyp (from_queue->stageset, to_queue)) !=
		TCML_COMPLETE) return (rc);
	if ((rc = nqs_movqtyp (from_queue->queuedset, to_queue)) !=
		TCML_COMPLETE) return (rc);
	if ((rc = nqs_movqtyp (from_queue->waitset, to_queue)) !=
		TCML_COMPLETE) return (rc);
	if ((rc = nqs_movqtyp (from_queue->holdset, to_queue)) !=
		TCML_COMPLETE) return (rc);
	switch (to_queue->q.type) {
	case QUE_BATCH:
		bsc_spawn();		/* Maybe spawn some batch reqs */
		break;
	case QUE_DEVICE:
		dsc_spawn();		/* Maybe spawn some device reqs */
		break;
	case QUE_PIPE:
		psc_spawn();		/* Maybe spawn some pipe reqs */
		break;
	}
	if (to_queue->q.status & QUE_UPDATE) {
		/*
		 *  No requests were spawned from the queue in which
		 *  the most recent request was placed.
		 */
		udb_qorder (to_queue);	/* Update and clear QUE_UPDATE bit */
	}
	return (rc);			/* Return transaction code */
}


/*** nqs_movquetyp
 *
 *
 *	long nqs_movquetyp():
 *	Move all of the NQS requests of a specific queue set to
 *	another queue.
 *
 *	Returns:
 *		TCML_COMPLETE:	if the request was successfully modified.
 *		TCML_NOSUCHQUE:	if the specified queue does not exist
 *				on this machine.
 *		TCML_QUEDISABL:	if the specified queue is disabled.
 *		TCML_NOSUCHREQ:	if the specified request does not exist
 *				on this machine.
 *		TCML_WROQUETYP:	if the destination queue type is not
 *				compatible with the source queue type.
 */
static long nqs_movqtyp (set, queue)
struct request *set;			/* Queue request set */
struct qhead *queue;			/* Queue to move requests to */
{
	register struct request *req;	/* Request struct allocated for req */
	register struct request *next;	/* Next request */
	register long rc;		/* return code from nqs_movreq() */
	long nqs_move();

	next = set;
	while ((req = next) != (struct request *) 0) {
		next = req->next;
		if (Debug > 2) {
			printf ("D$nqs_movqtyp(): Attempting to move request ");
			printf ("%ld\n", req->v1.req.orig_seqno);
			fflush (stdout);
		}
		rc = nqs_move (req->v1.req.orig_seqno,
			req->v1.req.orig_mid, queue);
		if (rc == TCML_COMPLETE) continue;
		return (rc);
	}
	return (TCML_COMPLETE);		/* Return transaction code */
}


/*** nqs_move
 *
 *
 *	long nqs_move():
 *	Move a queued NQS request.
 *
 *	Returns:
 *		TCML_COMPLETE:	if the request was successfully modified.
 *		TCML_NOSUCHREQ:	if the specified request does not exist
 *				on this machine.
 *		TCML_WROQUETYP:	if the destination queue type is not
 *				compatible with the source queue type.
 */
long nqs_move (orig_seqno, orig_mid, queue)
long orig_seqno;		/* Originating sequence number of request */
mid_t orig_mid;			/* Originating machine id */
register struct qhead *queue;	/* Queue in which req is placed */
{
	struct request *predecessor;	/* Predecessor in req set in queue */
	int state;			/* Request queue state: RQS_ */
	char path [MAX_PATHNAME+1];	/* Control file pathname */
	struct rawreq rawreq;		/* Raw request structure */
	int cfd;			/* Control file file-descriptor */
	register struct request *req;	/* Request struct allocated for req */		register struct qhead *from_queue;  /* Queue request originated in */
	register struct qcomplex *qcomplex; /* Queue complex pointer */
	int i;

	/*
	 *  Locate the request except for those that are departing, arriving,
	 *  or running.
	 */
	state = RQS_STAGING | RQS_QUEUED | RQS_WAITING | RQS_HOLDING |
		RQS_RUNNING;
	if ((req = nqs_fndreq (orig_seqno, orig_mid, &predecessor, &state)) ==
	    (struct request *) 0)
		return (TCML_NOSUCHREQ);	/* Request not found */
	else if (state & RQS_RUNNING)
		return (TCML_REQRUNNING);	/* Request already running */
	/*
	 *  Read the request header.
	 */
	if (Debug > 2) {
		printf ("D$nqs_move(): Read request (%s).\n", path);
		fflush (stdout);
	}
	if ((cfd = getreq ((long) orig_seqno, orig_mid, &rawreq)) == -1) {
		return (TCML_UNAFAILURE);	/* Failure */
	}
	/*
	 *  Verify that the request type and destination queue type agree
	 *  if the queue is a batch or device queue.
	 */
	if ((queue->q.type == QUE_BATCH && rawreq.type != RTYPE_BATCH) ||
	    (queue->q.type == QUE_DEVICE && rawreq.type != RTYPE_DEVICE)) {
		close (cfd);			/* Close request file */
		return (TCML_WROQUETYP);	/* wrong queue type */
	}
	/*
	 *  Copy the destination queue name into the queue field of the
	 *  rawreq structure for the request and rewrite the request header.
	 */
	if (Debug > 2) {
		printf ("D$nqs_move(): old queue name = %s, ",rawreq.quename);
		printf ("new queue name = %s\n", queue->q.namev.name);
		fflush (stdout);
	}
	strcpy (rawreq.quename, queue->q.namev.name);
	writereq (cfd, &rawreq);		/* Update request header */
	close (cfd);				/* Close request file */
	/*
	 *  The request has been located.  Remove the request from the
	 *  the containing queue and set the QUE_UPDATE bit.
	 */
	switch (state) {
	/*
	case RQS_RUNNING:
		if (predecessor == (struct request *) 0)
			req->queue->runset = req->next;
		else
			predecessor->next = req->next;
		req->queue->q.runcount--;
		break;
	*/
	case RQS_STAGING:
		if (predecessor == (struct request *) 0)
			req->queue->stageset = req->next;
		else
			predecessor->next = req->next;
		req->queue->q.stagecount--;
		break;
	case RQS_QUEUED:
		if (predecessor == (struct request *) 0)
			req->queue->queuedset = req->next;
		else
			 predecessor->next = req->next;
		req->queue->q.queuedcount--;
		break;
	case RQS_WAITING:
		if (predecessor == (struct request *) 0)
			req->queue->waitset = req->next;
		else
			predecessor->next = req->next;
		req->queue->q.waitcount--;
		break;
	case RQS_HOLDING:
		if (predecessor == (struct request *) 0)
			req->queue->holdset = req->next;
		else
			predecessor->next = req->next;
		req->queue->q.holdcount--;
		break;
	}
	req->next = (struct request *) 0;	/* No more requests */
	req->queue->q.status |= QUE_UPDATE;
	udb_qorder (req->queue);		/* Update queue ordering */
	req->queue = (struct qhead *) 0;
	/*
	 *  Now, schedule the request for execution.
	 */
	switch (queue->q.type) {
	case QUE_BATCH:
		req->v1.req.priority = bsc_sched (&rawreq);
		break;
	case QUE_DEVICE:
		req->v1.req.priority = dsc_sched (&rawreq);
		break;
	case QUE_PIPE:
		req->v1.req.priority = psc_sched (&rawreq);
		break;
	}
	if (queue->q.type == QUE_PIPE)
		/*
		 *  When a request is placed in a pipe queue, it is
		 *  routed and delivered to its destination as quickly
		 *  as possible, regardless of the -a time. The -a
		 *  time only becomes effective when the request has
		 *  reached its final destination batch or device queue.
		 */
		req->start_time = 0;
	else
		/*
		 *  The request is being placed in a batch or device
		 *  queue, and so the -a time parameter of the request
		 *  is now meaningful.
		 */
		req->start_time = rawreq.start_time;
					/* Remember start after time */
	/*
	 *  Place the prioritized request into the queue.
	 */
	if (state & RQS_RUNNING) {
		/*
		 *  Add the request to the running set for this queue
		 *  The QUE_UPDATE bit is set by a2s_a2rset().
		 */
		a2s_a2rset (req, queue);/* Add to running set */
	} else if (rawreq.flags & (RQF_OPERHOLD | RQF_USERHOLD)) {
		/*
		 *  Add the request to the hold set for this queue.
		 *  The QUE_UPDATE bit is set by a2s_a2hset().
		 */
		a2s_a2hset (req, queue);/* Add to hold set */
	} else if (rawreq.start_time > time ((time_t *) 0)) {
		/*
		 *  The request has a start time in the future.
		 *  The QUE_UPDATE bit is set by a2s_a2wset().
		 */
		a2s_a2wset (req, queue);/* Add to wait set */
	} else {
		/*
		 *  Place the request in the eligible to run set.
		 *  The QUE_UPDATE bit is set by a2s_a2qset().
		 */
		a2s_a2qset (req, queue);/* Add to queued set */
	}
	return (TCML_COMPLETE);
}



/*** nqs_holdreq
 *
 *
 *	long nqs_holdreq()
 *
 *	Place a hold on an NQS request.
 *
 *	Returns:
 *		TCML_COMPLETE:	if the request was successfully held.
 *		TCML_NOSUCHREQ:	if the specified request does not exist
 *				on this machine.
 *		TCML_NOTREQOWN:	if the mapped user-id does not match
 *				the current mapped user-id of the
 *				request owner.
 */
long nqs_holdreq (mapped_uid, orig_seqno, orig_mid)
uid_t mapped_uid;			/* Mapped owner user-id */
long orig_seqno;			/* Original request sequence# */
mid_t orig_mid;				/* Original machine-id of req */
{
	struct request *predecessor;	/* Predecessor in req set in queue */
	int state;			/* Request queue state: RQS_ */
	register struct qhead *queue;   /* Queue in which req is placed */
	register struct request *req;	/* Request struct allocated for req */
	int holdflag;

	/*
	 *  Locate the request.
	 */
	state = RQS_STAGING | RQS_QUEUED | RQS_WAITING | RQS_HOLDING |
		RQS_RUNNING;
	if ((req = nqs_fndreq (orig_seqno, orig_mid, &predecessor,
			       &state)) == (struct request *) 0) {
		/*
		 *  The request was not found in any of the local queues.
		 */
		return (TCML_NOSUCHREQ);
	} else if (state & RQS_RUNNING)
		return (TCML_REQRUNNING);	/* Request already running */
	if (mapped_uid && req->v1.req.uid != mapped_uid)
		/*
		 *  This request cannot be affected by the client.
		 */
		return (TCML_NOTREQOWN);
	queue = req->queue;
	switch (state) {
	case RQS_QUEUED:
		/*
		 *  The request is queued, so remove the request from the
		 *  queued set.
		 */
		if (predecessor == (struct request *) 0)
			queue->queuedset = req->next;
		else
			predecessor->next = req->next;
		queue->q.queuedcount--;		/* One less request queued */
		break;
	case RQS_WAITING:
		/*
		 *  The request is waiting to be started, so remove the request
		 *  from the waiting set.
		 */
		if (predecessor == (struct request *) 0)
			queue->waitset = req->next;
		else
			predecessor->next = req->next;
		queue->q.waitcount--;		/* One less request waiting */
		break;
	case RQS_HOLDING:
		/*
		 *  This request is already held.
		 */
		return (TCML_COMPLETE);
	}
	/*
	 *  Add the request to the holding set.
	 */
	a2s_a2hset (req, queue);
	/*
	 *  Set flags to indicate that the operator is holding this request.
	 */
	if (mapped_uid)
		holdflag = RQF_USERHOLD;
	else
		holdflag = RQF_OPERHOLD;
	if (flag_upd (req, holdflag, 0) != 0) return (TCML_UNAFAILURE);
	req->status |= holdflag;		/* Set Flag */
	/*
	 *  The NQS database queue image needs to be updated.
	 */
	udb_qorder (queue);
	return (TCML_COMPLETE);
}


/*** nqs_relreq
 *
 *
 *	long nqs_relreq()
 *
 *	Remove a hold on an queued NQS request.
 *
 *	Returns:
 *		TCML_COMPLETE:	if the request was successfully released.
 *		TCML_NOSUCHREQ:	if the specified request does not exist
 *				on this machine.
 *		TCML_NOTREQOWN:	if the mapped user-id does not match
 *				the current mapped user-id of the
 *				request owner.
 *		TCML_REQNOTHELD:if the specified request is not held.
 */
long nqs_relreq (mapped_uid, orig_seqno, orig_mid)
uid_t mapped_uid;			/* Mapped owner user-id */
long orig_seqno;			/* Original request sequence# */
mid_t orig_mid;				/* Original machine-id of req */
{
	struct request *predecessor;	/* Predecessor in req set in queue */
	int state;			/* Request queue state: RQS_ */
	register struct qhead *queue;   /* Queue in which req is placed */
	register struct request *req;	/* Request struct allocated for req */

	/*
	 *  Locate the request.
	 */
	state = RQS_STAGING | RQS_QUEUED | RQS_WAITING | RQS_HOLDING |
		RQS_RUNNING;
	if ((req = nqs_fndreq (orig_seqno, orig_mid, &predecessor,
			       &state)) == (struct request *) 0)
		/*
		 *  The request was not found in any of the local queues.
		 */
		return (TCML_NOSUCHREQ);
	if (mapped_uid && req->v1.req.uid != mapped_uid)
		/*
		 *  This request cannot be affected by the client.
		 */
		return (TCML_NOTREQOWN);
	if (!(state & RQS_HOLDING))
		/*
		 *  This request is not on hold.
		 */
		return (TCML_REQNOTHELD);
	if (mapped_uid && (req->status & RQF_OPERHOLD))
		/*
		 *  This request cannot be affected by the client (hold
		 *  by NQS operator).
		 */
		return (TCML_INSUFFPRV);
	/*
	 *  Clear RQF_OPERHOLD.
	 */
	if (flag_upd (req, 0, RQF_OPERHOLD | RQF_USERHOLD) != 0)
		return (TCML_UNAFAILURE);
	req->status &= ~(RQF_OPERHOLD | RQF_USERHOLD);
						/* Clear Flag */
	queue = req->queue;
	/*
	 *  Remove the request to the holding set.
	 */
	if (predecessor == (struct request *) 0)
		queue->holdset = req->next;
	else
		predecessor->next = req->next;
	queue->q.holdcount--;			/* One less request held */
	if (req->start_time > time ((time_t *) 0)) {
		/*
		 *  The request has a future start time, so add the request to
		 *  the waiting set.
		 */
		a2s_a2wset (req, queue);
	} else {
		/*
		 *  Add the request to the queued set.
		 */
		a2s_a2qset (req, queue);
		/*
		 *  Maybe spawn a request.
		 */
		switch (queue->q.type) {
		case QUE_BATCH:
			bsc_spawn ();
			break;
		case QUE_DEVICE:
			dsc_spawn ();
			break;
		case QUE_PIPE:
			psc_spawn ();
			break;
		}
	}
	/*
	 *  The NQS database queue image needs to be updated.
	 */
	udb_qorder (queue);
	return (TCML_COMPLETE);
}


/***    flag_upd
 *
 *
 *	int flag_upd()
 *
 *	Update the flags field in a raw request in a control file
 *
 *	Returns:
 *		 0:	if successful.
 *		-1:	if failure.
 */
static int flag_upd (req, setflags, clflags)
struct request *req;			/* Ptr to request structure */
short setflags;				/* Flags to set */
short clflags;				/* Flags to clear */
{
	int cfd;			/* File descriptor for control file */
	struct rawreq rawreq;		/* Raw request structure */

	/*
	 *   Locate and read raw request header.
	 */
	if ((cfd = gethdr (req->v1.req.orig_seqno, req->v1.req.orig_mid,
	     &rawreq)) == -1 ) {
		printf ("E$flag_upd(): Read of request header failed.");
		fflush (stdout);
		return (-1);
	}
	/*
	 *  Update flags.
	 */
	rawreq.flags |= setflags;
	rawreq.flags &= ~clflags;
	/*
	 *  Rewrite raw request to control file.
	 */
	if (writehdr (cfd, &rawreq) == -1) {
		printf ("E$flag_upd(): Write of request header failed.");
		fflush (stdout);
		close (cfd);
		return (-1);
	}
	close (cfd);
	return (0);
}


/*** nqs_rerunreq
 *
 *
 *	long nqs_rerunreq():
 *
 *	Rerun a running NQS request (PKT_RERUNREQ).
 *
 *	Returns:
 *		TCML_COMPLETE:	 if the request was requeued for rerun.
 *		TCML_NOSUCHREQ:	 if the specified request does not exist on
 *				 this machine.
 *		TCML_WROQUETYP:	 if the specified request is not of type BATCH.
 *		TCML_REQNOTRNNG: if the specified request is not running.
 */
long nqs_rerunreq (orig_seqno, orig_mid)
long orig_seqno;			/* Original request sequence# */
mid_t orig_mid;				/* Original machine-id of req */
{
	char path [MAX_PATHNAME+1];	/* Pathname of script link file */
	struct qhead *queue;		/* Queue in which request resides */
	struct request *predecessor;	/* Predecessor in req set in queue */
	int cfd;			/* File descriptor for control file */
	int state;			/* Request queue state: RQS_ */
	register struct request *req;	/* Request struct allocated for req */
	struct rawreq rawreq;		/* Raw request structure */
	struct running *runvar;		/* Runvars entry for req being rerun */
	short reqindex;			/* Request index in Runvars if */
					/* the request is running, or */
					/* index in Runvars if subrequest */
					/* is transporting parent request */
					/* to its destination via a pipe/ */
					/* network queue pair */
	struct transact transaction;	/* Transaction for request */
	register struct qcomplex *qcomplex; /* Queue complex pointer */
	int i;

	/*
	 *  Locate the request.
	 */
	state = RQS_STAGING | RQS_QUEUED | RQS_WAITING | RQS_HOLDING
		| RQS_ARRIVING | RQS_RUNNING;
	if ((req = nqs_fndreq (orig_seqno, orig_mid, &predecessor, &state)) ==
					(struct request *) 0)
		/*
		 *  The request was not found in any of the local queues.
		 */
		return (TCML_NOSUCHREQ);
	if (req->queue->q.type != QUE_BATCH)
		/*
		 *  This is not a batch request.
		 */
		return (TCML_WROQUETYP);
	if (!(req->status & RQF_RESTARTABLE))
		/*
		 *  This request can not be rerun.
		 */
		return (TCML_NORERUN);
	if (state != RQS_RUNNING)
		/*
		 *  This batch request is not running.
		 */
		return (TCML_REQNOTRNNG);
	queue = req->queue;		/* Queue this request resides in */
	nqs_deque (req);
	for (i = MAX_COMPLXSPERQ; --i >= 0;) {
		qcomplex = queue->v1.batch.qcomplex[i];
		if (qcomplex == (struct qcomplex *)0) continue;
		qcomplex->runcount--;
	}
	Gblbatcount--;			/* One less batch request */
	transaction.state = RTS_STASIS;
	if (tra_setstate (req->v1.req.trans_id, &transaction) == -1) {
		/*
		 *  An error occurred modifying the request
		 *  transaction state.
		 */
		printf ("E$Error setting transaction state ");
		printf ("in nqs_rerunreq().\n");
		printf ("I$Transaction-id: %1d.\n", req->v1.req.trans_id);
		fflush (stdout);
	}
	/*
	 *  Remove the request's script (data file) link file. This file will
	 *  only exist if the shell strategy is free. Delete also all request
	 *  output files.
	 */
	if ((cfd = getreq ((long) req->v1.req.orig_seqno,
				req->v1.req.orig_mid, &rawreq)) == -1 ) {
		printf ("E$nqs_rerunreq(): Read of raw request failed.\n");
		return (TCML_UNAFAILURE);
	}
	pack6name (path, Nqs_scripts, (char *) 0,
			(char *) 0, (long) rawreq.orig_seqno, 5,
			(long) rawreq.orig_mid, 6, 0, 3);
	unlink (path);
	unlink (namstderr (orig_seqno, orig_mid));
	unlink (namstdlog (orig_seqno, orig_mid));
	unlink (namstdout (orig_seqno, orig_mid));
	close (cfd);			/* Close request file */
	/*
	 *  If the user has indicated that the request could be rerun, then
	 *  notify its shepherd to requeue the request upon termination.
	 */
	reqindex = req->reqindex;
#if	SGI | SYS52 | SYS53 | UTS | UNICOS
	kill (-(Runvars + reqindex)->process_family, SIGKILL);
#else
#if	BSD42 | BSD43 | ULTRIX
	killpg((Runvars + reqindex)->process_family, SIGKILL);
#else
BAD SYSTEM TYPE
#endif
#endif
	/*
	 *  Add the request to the waiting set and maybe it can be
	 *  spawned as well. At the point there is a symchronization issue
	 *  between the request dying and being re-started. So the queued
	 *  set can not be used!
	 */
	req->start_time = time ((time_t *) 0) + MDELTA;
	a2s_a2wset (req, queue);
	bsc_spawn ();
	if (queue->q.status & QUE_UPDATE) {
		/*
		 *  No requests were spawned from the queue in which
		 *  the most recent request was placed.
		 */
		udb_qorder (queue);	/* Update and clear QUE_UPDATE bit */
	}
	return (TCML_COMPLETE);
}
