/*++ nqs_wakreq.c - Network Queueing System
 *
 * $Source$
 *
 * DESCRIPTION:
 *
 *	Wakeup requests that are being held by -a times which have now
 *	occurred.
 *
 *
 *	Author:
 *	-------
 *	Brent A. Kingsbury, Sterling Software Incorporated.
 *	August 12, 1985.
 *
 *
 * STANDARDS VIOLATIONS:
 *   None.
 *
 * REVISION HISTORY: ($Revision$ $Date$ $State$)
 * $Log$
 *
 *	November 11, 1991 Reduce the number of alarms in load-balancing queues
 *
 */

#include "nqs.h"
#include "nqsxvars.h"			/* Global vars and directories */

extern void a2s_a2qset();		/* Add request to queued set */
extern void bsc_spawn();		/* Maybe spawn some batch reqs */
extern void dsc_spawn();		/* Maybe spawn some device reqs */
extern void nqs_vtimer();		/* Set virtual timer */
extern void psc_spawn();		/* Maybe spawn some pipe reqs */
extern time_t time();			/* GMT time since 0:00:00 Jan 1, 1970 */
extern void udb_qorder();		/* Update NQS database queue image */

/*** nqs_wakreq
 *
 *
 *	void nqs_wakreq():
 *
 *	Move all requests in all queue waiting sets whose -a times have
 *	occurred, to the eligible to run request sets in their respective
 *	queues.
 */
void nqs_wakreq()
{
	void wakeupset();		/* Wakeup requests in a queue set */

	wakeupset (Nonnet_queueset);	/* Wakeup requests in non-network */
					/* queues */
	wakeupset (Net_queueset);	/* Wakeup requests in network queues*/
}


/*** wakeupset
 *
 *
 *	void wakeupset():
 *
 *	Move all requests in all waiting sets whose -a times have occurred,
 *	to the eligible to run request sets in their respective queues, as
 *	present in the given queue set.
 */
static void wakeupset (queueset)
struct qhead *queueset;			/* Queue set */
{
	register struct qhead *queue;
	register struct request *prevreq;
	register struct request *req;
	time_t timenow;			/* Current GMT time */

	time (&timenow);		/* Get current time */
	queue = queueset;
	while (queue != (struct qhead *) 0) {
		prevreq = (struct request *)0;
		req = queue->waitset;
		while (req != (struct request *)0) {
			if (req->start_time <= timenow) {
				/*
				 *  This request is now eligible for execution.
				 *  Remove the request from the waiting set and
				 *  place it in the eligible queue set.
				 */
				if (prevreq == (struct request *)0) {
					queue->waitset = req->next;
				}
				else prevreq->next = req->next;
				queue->q.waitcount--;
				/*
				 *  Add the request to the eligible to run
				 *  set.
				 */
				a2s_a2qset (req, queue);
				/*
				 *  Continue scanning the waiting set of
				 *  requests for the queue.
				 */
				if (prevreq == (struct request *)0) {
					req = queue->waitset;
				}
				else req = prevreq->next;
			}
			else {
				/*
				 *  This request has a start time in the
				 *  future.  Set the virtual alarm to wake
				 *  it up later.
				 */
				nqs_vtimer (&req->start_time, nqs_wakreq);
				if (prevreq == (struct request *)0 &&
				    strncmp (CPIPQUENAM,
					    queue->q.namev.name,
					    strlen (CPIPQUENAM)) == 0)
					/*
					 *  In load-balancing queues, we set
					 *  only the virtual alarm on the
					 *  first request waiting in the list.
					 */
					break;
				else prevreq = req;
				req = req->next;
			}
		}
		queue = queue->next;
	}
	/*
	 *  Request(s) may be eligible for spawning.
	 */
	bsc_spawn();			/* Maybe spawn batch reqs */
	dsc_spawn();			/* Maybe spawn device reqs */
	psc_spawn();			/* Maybe spawn pipe reqs */
	/*
	 *  Update queue states whose images are out of date in the NQS
	 *  database.
	 */
	queue = queueset;
	while (queue != (struct qhead *)0) {
		if (queue->q.status & QUE_UPDATE) {
			/*
			 *  Update queue image and clear the QUE_UPDATE
			 *  bit.
			 */
			udb_qorder (queue);
		}
		queue = queue->next;
	}
}
