/*++ qhold.c - Network Queueing System
 *
 * $Source$
 *
 * DESCRIPTION:
 *
 *	Hold a queued or waiting request.
 *
 *
 *	Author:
 *	-------
 *	Christian Boissat, CERN, Geneva, Switzerland.
 *	January 6, 1992.
 *
 *
 * STANDARDS VIOLATIONS:
 *   None.
 *
 * REVISION HISTORY: ($Revision$ $Date$ $State$)
 * $Log$
 *
 */

#define	MAX_REQS	100		/* Maximum number of reqs that */
					/* can be held at a time. */
#include <stdio.h>
#include <signal.h>			/* Signal definitions */
#include <pwd.h>			/* Password file definitions */
#include "nqs.h"			/* NQS types and definitions */
#include NETDB				/* Network database header file; */
					/* Proto/Makefile will tell us */
					/* where to find netdb.h */
#include "nqsdirs.h"			/* NQS files and directories */
#include "transactcc.h"			/* Transaction completion codes */


/*
 *	External functions:
 */
extern long hdlreq();			/* Hold request */
extern void closepwdb();		/* Close account/password database */
extern void exiting();			/* Exiting.  Relinquish inter- */
					/* process communication file */
extern struct passwd *fetchpwnam();	/* Fetch passwd entry by name */
extern uid_t getuid();			/* Get real user-id */
#if	SGI | SYS52 | SYS53 | ULTRIX | UNICOS | UTS
#else
#if	BSD42 | BSD43
extern int interconn();			/* Make up for the lack of named */
					/* pipes */
#else
BAD SYSTEM TYPE
#endif
#endif
extern int reqspec();			/* Parse request-id returning */
					/* completion code, request-seq#, */
					/* and req machine-id */
extern int localmid();			/* Get local machine-id */
extern int nqspriv();			/* Return non-zero if NQS mgr */

/*
 *	Global variables:
 */
char *Qhold_prefix = "Qhold";

/*** main
 *
 *
 *	qhold [ -u <username> ] <request-id(s)>
 */
main (argc, argv)
int argc;
char **argv;
{
	void cleanup();		/* Cleanup on SIGINT, etc. */
	void diaghold();	/* Diagnose hdlreq completion code */
	void showhow();		/* Show how to use this command */

	int n_reqs;			/* Number of reqs to hold. */
	struct {
		long orig_seqno;	/* Sequence# for req */
		mid_t machine_id;	/* Machine-id of originating machine */
	} reqs [MAX_REQS];		/* Reqs to hold */
	struct passwd *passwd;		/* Password structure ptr */
	char **scan_reqs;		/* Scan reqs */
	char *argument;			/* Ptr to cmd line arg text */
	uid_t real_uid;			/* Real user-id */
	mid_t local_mid;		/* local machine-id */


	/*
	 *  Catch 4 common household signals:  SIGINT, SIGQUIT,SIGHUP, 
	 *  and SIGTERM.  This is quite important, because we do not want
	 *  to leave useless inter-process communication files hanging
	 *  around in the NQS directory hierarchy.  However, if we do,
	 *  it is not fatal, it will eventually be cleaned up by NQS.
	 */
	signal (SIGINT, cleanup);
	signal (SIGQUIT, cleanup);
	signal (SIGHUP, cleanup);
	signal (SIGTERM, cleanup);
	if (chdir (Nqs_root) == -1) {
		fprintf (stderr, "%s(FATAL): Unable to chdir() to the NQS ",
			 Qhold_prefix);
		fprintf (stderr, "root directory.\n");
		exit (1);
	}
	/*
	 *  On systems with named pipes, we get a pipe to the local
	 *  daemon automatically the first time we call inter().
	 */
#if	SGI | SYS52 | SYS53 | ULTRIX | UNICOS | UTS
#else
#if	BSD42 | BSD43
	if (interconn () < 0) {
		fprintf (stderr, "%s(FATAL): Unable to get ", Qhold_prefix);
		fprintf (stderr, "a pipe to the local daemon.\n");
		exit (1);
	}
#else
BAD SYSTEM TYPE
#endif
#endif
	passwd = NULL;			/* No -u flag seen */
	while (*++argv != NULL && **argv == '-') {
		argument = *argv;
		if (*++argument == 'u') { /* User-name specification */
			if (*++argv == NULL) {
				fprintf (stderr, "Missing username.\n");
				exit (1);
			}
			if (passwd != NULL) {
				fprintf (stderr,
					"Multiple -u specifications.\n");
				exit (1);
			}
			if ((passwd = fetchpwnam (*argv)) == NULL) {
				fprintf (stderr, "No such user on this ");
				fprintf (stderr, "machine.\n");
				exit (1);
			}
			if (localmid (&local_mid) != 0) {
				fprintf (stderr, "%s(FATAL): ", Qhold_prefix);
				fprintf (stderr, "Unable to get machine-id");
				fprintf (stderr, "of local host.\n");
				exit(1);
			}
			if ((nqspriv (getuid(), local_mid, local_mid)
				& QMGR_OPER_PRIV) ||
			    passwd->pw_uid == getuid()) {
				/*
				 *  We have NQS operator privileges, or we
				 *  are just going after our own requests.
				 */
				real_uid = passwd->pw_uid;
			}
			else {
				fprintf (stderr, "Insufficient privilege ");
				fprintf (stderr, "for -u specification.\n");
				exit (1);
			}
		} else {
			fprintf (stderr, "Invalid option flag ");
			fprintf (stderr, "specified.\n");
			showhow();
		}
	}
	if (passwd == NULL) {
		/*
		 *  No username specified.  We assume the invoker.
		 */
		real_uid = getuid();		/* Get real user-id */
	}
	else closepwdb();		/* Close account/password database */
	/*
	 *  Build the set of reqs to be held.
	 */
	if (*argv == NULL) {
		/*
		 *  No request-ids were specified.
		 */
		fprintf (stderr, "No request-id(s) specified.\n");
		showhow();
	}
	else {
		n_reqs = 0;			/* #of reqs to hold */
		scan_reqs = argv;		/* Set req scan pointer */
		while (*scan_reqs != NULL &&	/* Loop to hold reqs */
		       n_reqs < MAX_REQS) {
			switch (reqspec (*scan_reqs, &reqs [n_reqs].orig_seqno,
					 &reqs [n_reqs].machine_id)) {
			case -1:
				fprintf (stderr, "Invalid request-id syntax ");
				fprintf (stderr, "for request-id: %s.\n",
					*scan_reqs);
				exit (1);
			case -2:
				fprintf (stderr, "Unknown machine for");
				fprintf (stderr, "request-id: %s.\n",
					*scan_reqs);
				exit (1);
			case -3:
				fprintf (stderr, "Network mapping database ");
				fprintf (stderr, "inaccessible.  Seek staff ");
				fprintf (stderr, "support.\n");
				exit (1);
			case -4:
				fprintf (stderr, "Network mapping database ");
				fprintf (stderr, "error when parsing ");
				fprintf (stderr, "request-id: %s.\n",
					*scan_reqs);
				fprintf (stderr, "Seek staff support.\n");
				exit (1);
			}
			scan_reqs++;		/* One more req */
			n_reqs++;
		}
		if (*scan_reqs != NULL) {
			/*
			 *  Too many reqs were specified to be held.
			 */
			fprintf (stderr, "Too many requests given to ");
			fprintf (stderr, "hold.\n");
			exit (1);
		}
		/*
		 *  Now that everything has been parsed and legitimized,
		 *  hold the specified set of requests.
		 */
		n_reqs = 0;
		while (*argv != NULL) {		/* Loop to hold reqs */
			diaghold (hdlreq (0, real_uid, reqs [n_reqs].orig_seqno,
				 reqs [n_reqs].machine_id), *argv);
			argv++;			/* One more req */
			n_reqs++;
		}
	}
	exiting();			/* Delete our comm. file */
	exit (0);
}


/*** cleanup
 *
 *
 *	Catch certain signals, and delete the inter-process
 *	communication file we have been using.
 */
static void cleanup (sig)
int sig;				/* Signal that was caught */
{
	signal (sig, SIG_IGN);		/* Ignore multiple signals */
	exiting();			/* Delete our comm. file */
}


/*** diaghold
 *
 *
 *	diaghold():
 *	Diagnose hdlreq() completion code.
 */
static void diaghold (code, reqid)
long code;				/* Transaction completion code */
char *reqid;				/* Character version of request-id */
{
	switch (code) {
	case TCML_INTERNERR:
		printf ("Internal error.\n");
		exiting();		/* Delete communication file */
		exit (1);
	case TCML_NOESTABLSH:
		printf ("Unable to establish inter-process communications ");
		printf ("with NQS daemon.\n");
		printf ("Seek staff support.\n");
		exiting();		/* Delete communication file */
		exit (1);
	case TCML_NOLOCALDAE:
		printf ("The NQS daemon is not running.\n");
		printf ("Seek staff support.\n");
		exiting();		/* Delete communication file */
		exit (1);
	case TCML_NOSUCHREQ:
		printf ("Request %s does not exist.\n", reqid);
		break;
	case TCML_NOTREQOWN:
		printf ("Not owner of request %s.\n", reqid);
		break;
	case TCML_PEERDEPART:
		printf ("Request %s is presently being routed by a ", reqid);
		printf ("pipe queue,\n");
		printf ("and this NQS implementation does not support the ");
		printf ("handling of a\n");
		printf ("request under such conditions.\n");
		break;
	case TCML_PROTOFAIL:
		printf ("Protocol failure in inter-process communications ");
		printf ("with NQS daemon.\n");
		printf ("Seek staff support.\n");
		exiting();		/* Delete communication file */
		exit (1);
	case TCML_COMPLETE:
		printf ("Request %s has been held.\n", reqid);
		break;
	case TCML_REQRUNNING:
		printf ("Request %s is running.\n", reqid);
		break;
	default:
		printf ("Unexpected completion code from NQS daemon.\n");
		exiting();		/* Delete communication file */
		exit (1);
	}
}


/*** showhow
 *
 *
 *	showhow():
 *	Show how to use this command.
 */
static void showhow()
{
	fprintf (stderr, "Command format is:\n\n");
	fprintf (stderr, "    qhold [ -u <username> ] <request-id(s)>\n\n");
	exit (0);
}
